import os
import os.path
import atexit
import rox
from rox import g
import gobject
import dbus
import dbus.service
import rox.options
import rox.choices
import rox.OptionsBox
import rox.InfoWin
import xdg.Config
import xdg.Menu
from i18n import _
from config import *
from ctxmenu import CtxMenuWidget
from imagemenu import ImageMenuItem
from xdgmenu import XdgMenu, XdgMenuItem
import roxmenu
from roxmenu import RoxMenu, RoxMenuItem

class OptionDefaultCallable:
  """This mixed class allows to pass function as
  default value. When default_value is requested from
  rox.options.Option this function will be called and
  result is assumed as default value for this option.
  """

  def __init__(self, class_obj, name, value, group = None):
    self._callable_class = class_obj
    if callable(value): v = ""
    else: v = value
    class_obj.__init__(self, name, v, group)
    if callable(value):
      self._callable_defaut_value = value
      del(self.default_value)

  def __getattr__(self, name):
    if name == "default_value":
      return self._callable_defaut_value()
    if hasattr(self._callable_class, "__getattr__"):
      return self._callable_class.__getattr__(self, name)
    raise AttributeError(name)

class MyOption(rox.options.Option, OptionDefaultCallable):
  """Option with default value callable capability."""

  def __init__(self, name, value, group = None):
    OptionDefaultCallable.__init__(self,
        rox.options.Option, name, value, group)

class MyListOption(rox.options.ListOption, OptionDefaultCallable):
  """ListOption with default value callable capability."""

  def __init__(self, name, value, group = None):
    OptionDefaultCallable.__init__(self,
        rox.options.ListOption, name, value, group)

def check_programs(progs, fallback):
  """Check list of programs progs in directories specified
  in the PATH environment variable. Return first occurrence
  or fallback."""
  try:
    path = os.getenv("PATH").split(":")
  except AttributeError: return fallback
  for p in progs:
    for d in path:
      if os.access(os.path.join(d, p), os.X_OK): return p
  return fallback

def default_terminal():
  """Compute default terminal emulator."""
  return check_programs(("gnome-terminal", "xterm",
    "rxvt", "aterm", "eterm"), "xterm")

def default_uri_handler():
  """Compute default URI handler."""
  return check_programs(("firefox", "mozilla",
    "galeon", "dillo", "opera", "netscape",
    "htmlview"), "mozilla")

def default_wrapper_dir():
  """Compute default wrappers directory."""
  return os.path.join(os.path.expanduser("~"),
      ".cache", app_name, "wrappers")

def app_dir_def_2_str(d):
  """Convert dictionary that represent applications directory
  definition to string format."""
  return ":".join([d["menu_base"], d["directory"],
    str(int(d["recursive"]))])

def str_2_app_dir_def(s):
  """Convert string that represent applications directory
  definition to dictionary format."""
  b, d, r = s.split(":")
  return {"menu_base":b, "directory":d, "recursive":bool(int(r))}

def default_app_dir_defs():
  """Compute default applications directory definitions."""
  defs = []
  def dir_in_defs(dir):
    for v in defs:
      if v["directory"] == dir:
        return v
    return None
  try: realpath = os.path.realpath
  except AttributeError: realpath = lambda x: x
  defs.append(roxmenu.dir_def("",
    realpath(os.path.dirname(rox.app_dir)), False))
  for prefix in ("/usr/lib", "/usr/local/lib", "/opt",
      os.path.expanduser("~")):
    check_prefix = True
    for subdir in ("ROX-Apps", "rox-apps", "Apps", "apps"):
      dir = os.path.join(prefix, subdir)
      if not os.path.isdir(dir): continue
      dir = realpath(dir)
      d = dir_in_defs(dir)
      if d: d["recursive"] = True
      else: defs.append(roxmenu.dir_def("", dir, True))
      check_prefix = False
    if not check_prefix: continue
    if not os.path.isdir(prefix): continue
    dir = realpath(prefix)
    if dir_in_defs(dir): continue
    defs.append(roxmenu.dir_def("", dir, False))
  return map(app_dir_def_2_str, defs)

def is_valid_xdg_menu_file(filename, menu=None):
  if not filename:
    return False
  if menu is None:
    try:
      menu = xdg.Menu.parse(filename)
    except:
      return False
  if os.path.isabs(filename):
    return menu.Filename == filename
  return menu.Filename.endswith(filename)

def default_include_xdg_menus():
  result = []
  for filename in (
          'preferences.menu',
          'system-settings.menu',
          'gnome-settings.menu'):
    if is_valid_xdg_menu_file(filename):
      result.append(filename)
  if not result and is_valid_xdg_menu_file('settings.menu'):
    result.append('settings.menu')
  return ", ".join(result)

rox.choices.migrate(app_name, site_name)
rox.setup_app_options(app_name, site=site_name)
icon_size = MyOption("icon_size", "menu")
terminal = MyOption("terminal", default_terminal)
uri_handler = MyOption("uri_handler", default_uri_handler)
preload_menu = MyOption("preload_menu", 1)
preload_icons = MyOption("preload_icons", 0)
show_hidden = MyOption("show_hidden", 0)
show_missing = MyOption("show_missing", 1)
applet_image = MyOption("applet_image", "")
wrapper_dir = MyOption("wrapper_dir", default_wrapper_dir)
force_utf8_wrappers = MyOption("force_utf8_wrappers", 1)
app_dir_defs = MyListOption("app_dir_defs", default_app_dir_defs)
enable_rox_apps = MyOption("enable_rox_apps", 1)
enable_logout = MyOption("enable_logout", 1)
include_xdg_menus = MyOption("include_xdg_menus",
    default_include_xdg_menus)
rox.app_options.notify()

def build_app_dir_defs(box, node, label, option):
  """Build applications directory definitions control for
  rox.OptionsBox."""
  cont=rox.g.VBox(False, 4)
  if label:
    label_wid = g.Label(label)
    cont.pack_start(label_wid, False, True, 0)
    label_wid.show()
  swin = g.ScrolledWindow()
  swin.set_border_width(4)
  swin.set_policy(g.POLICY_NEVER, g.POLICY_ALWAYS)
  swin.set_shadow_type(g.SHADOW_IN)
  swin.set_size_request(-1, 128)
  cont.pack_start(swin, True, True, 0)
  model = g.ListStore(str, str, bool)
  view = g.TreeView(model)
  swin.add(view)
  selection=view.get_selection()
  selection.set_mode(g.SELECTION_SINGLE)
  view.set_reorderable(True)
  def cell_edited(ell, path, new_text, col):
    if col in (0, 1) and new_text.find(':')>=0:
      return
    iter=model.get_iter_from_string(path)
    model.set(iter, col, new_text)
    box.check_widget(option)
  def cell_toggled(ell, path, col):
    iter=model.get_iter_from_string(path)
    model.set(iter, col, not model.get_value(iter, col))
    box.check_widget(option)
  cell = g.CellRendererText()
  column = g.TreeViewColumn(_("Menu Prefix"), cell, text = 0)
  view.append_column(column)
  cell.set_property('editable', True)
  cell.connect('edited', cell_edited, 0)
  cell = g.CellRendererText()
  column = g.TreeViewColumn(_("Directory"), cell, text = 1)
  view.append_column(column)
  cell.set_property('editable', True)
  cell.connect('edited', cell_edited, 1)
  cell = g.CellRendererToggle()
  column = g.TreeViewColumn(_("Recursive"), cell, active = 2)
  view.append_column(column)
  cell.set_property("activatable", True)
  cell.connect('toggled', cell_toggled, 2)
  def add(widget, box):
    iter=model.append()
    model.set(iter, 0, "", 1, "", 2, False)
    view.get_selection().select_iter(iter)
    box.check_widget(option)
  def remove(widget, box):
    ign, iter = view.get_selection().get_selected()
    if iter: model.remove(iter)
    box.check_widget(option)
  hbox=g.HBox(False, 2)
  cont.pack_start(hbox, False)
  but=g.Button(stock=g.STOCK_ADD)
  but.connect('clicked', add, box)
  hbox.pack_start(but, False)            
  but=g.Button(stock=g.STOCK_REMOVE)
  but.connect('clicked', remove, box)
  hbox.pack_start(but, False)            
  box.may_add_tip(swin, node)
  def get():
    v=[]
    iter=model.get_iter_first()
    while iter:
      b=model.get_value(iter, 0)
      d=model.get_value(iter, 1)
      r=model.get_value(iter, 2)
      v.append(app_dir_def_2_str(roxmenu.dir_def(b,d,r)))
      iter=model.iter_next(iter)
    return v
  def set():
    model.clear()
    for v in option.list_value:
      d = str_2_app_dir_def(v)
      iter=model.append()
      model.set(iter, 0, d["menu_base"], 1,
          d["directory"], 2, d["recursive"])
  box.handlers[option]=(get, set)
  return [cont]

rox.OptionsBox.widget_registry["app_dir_defs"] = build_app_dir_defs

icon_size_tr = {
  "disable": g.ICON_SIZE_INVALID,
  "menu": g.ICON_SIZE_MENU,
  "small_toolbar": g.ICON_SIZE_SMALL_TOOLBAR,
  "large_toolbar": g.ICON_SIZE_LARGE_TOOLBAR,
  "button": g.ICON_SIZE_BUTTON,
  "dnd": g.ICON_SIZE_DND,
  "dialog": g.ICON_SIZE_DIALOG,
}

class Button(g.EventBox, CtxMenuWidget):
  """XDG-Menu Button."""

  class DBusObject(dbus.service.Object):
    def __init__(self, button, bus, path):
      dbus.service.Object.__init__(self, bus, path)
      self.button = button

    @dbus.service.method(dbus_button_iface, in_signature='', out_signature='')
    def Refresh(self):
      self.button.refresh()

    @dbus.service.method(dbus_button_iface, in_signature='', out_signature='')
    def PopupStartMenu(self):
      self.button.popup_start_menu()

    @dbus.service.method(dbus_button_iface, in_signature='', out_signature='')
    def PopupStartMenuAtPointer(self):
      self.button.popup_start_menu_at_pointer()

    @dbus.service.method(dbus_button_iface, in_signature='', out_signature='')
    def Quit(self):
      self.button.quit()

  def __init__(self):
    g.EventBox.__init__(self)
    xdg.Config.setWindowManager("ROX")
    xdg.Config.setIconTheme(
        g.settings_get_default().get_property("gtk-icon-theme-name"))
    self.set_size_request(-1, -1)
    self.image = None
    self._setup_image()
    self._xmenu = {}
    self._rox_tree = None
    self._start_menu = None
    self.connect("button-press-event", self._button_press)
    self.connect("size-allocate", self._size_alloc)
    self.ctx_menu_position = self.position_menu
    self.ctx_menu_enable()
    rox.app_options.add_notify(self._options_changed)
    if preload_menu.int_value:
      self.get_start_menu()
    self._dbus_object = self.DBusObject(self, dbus.SessionBus(), '/Button')

  def _get_image_pixbuf(self, w = None, h = None):
    path = applet_image.value
    def_path = os.path.join(rox.app_dir, ".DirIcon")
    if not path:
      path = def_path
    elif not os.path.isabs(path) and os.path.sep not in path:
      if w and h:
        size = (w + h) / 2
      else:
        size = None
      path = xdg.IconTheme.getIconPath(path, size)
      if not path:
        path = def_path
    while True:
      try:
        if w and h:
          pixbuf = g.gdk.pixbuf_new_from_file_at_size(
              path, w - 2, h - 2)
        else:
          pixbuf = g.gdk.pixbuf_new_from_file(path)
      except gobject.GError:
        pixbuf = None
      if pixbuf is None and path != def_path:
        path = def_path
        continue
      break
    return pixbuf

  def _setup_image(self):
    pixbuf = self._get_image_pixbuf()
    w = pixbuf.get_width()
    h = pixbuf.get_height()
    if self.image is not None:
      self.image.destroy()
    self.image = g.Image()
    self.image_w = 0
    self.image_h = 0
    self.image_aspect = float(w) / h
    self.add(self.image)
    self.image.show()

  def _resize(self, w, h, vertical=None):
    if vertical is None: vertical = self.is_vertical()
    if w != self.image_w or h != self.image_h:
      self.image.set_from_pixbuf(
          self._get_image_pixbuf(w, h))
      self.image_w = w
      self.image_h = h
      if vertical:
        self.image.set_size_request(8, h)
      else:
        self.image.set_size_request(w, 8)

  def _size_alloc(self, widget, a):
    if self.is_vertical():
      h = int(a.width / self.image_aspect)
      self._resize(a.width, h, True)
    else:
      w = int(a.height * self.image_aspect)
      self._resize(w, a.height, False)

  def ctx_menu_fill(self, m):
    a = self.ctx_menu_append
    a(m, g.ImageMenuItem(g.STOCK_REFRESH), self.refresh)
    a(m, g.ImageMenuItem(g.STOCK_PREFERENCES), lambda i: rox.edit_options())
    a(m, g.ImageMenuItem(g.STOCK_DIALOG_INFO),
        lambda i: rox.InfoWin.infowin(app_name).show())
    a(m, g.ImageMenuItem(g.STOCK_REMOVE), self.quit)

  def get_xmenu(self, filename=''):
    if filename not in self._xmenu:
      try:
        self._xmenu[filename] = xdg.Menu.parse(filename or None)
      except:
        return None
      if filename and not is_valid_xdg_menu_file(
          filename, self._xmenu[filename]):
        return None
    return self._xmenu[filename]

  def get_rox_tree(self):
    if self._rox_tree is None:
      defs = map(str_2_app_dir_def, app_dir_defs.list_value)
      self._rox_tree = roxmenu.get_menu_tree(defs)
    return self._rox_tree

  def get_start_menu(self):
    if self._start_menu is None:
      try: size = icon_size_tr[icon_size.value]
      except KeyError: size = g.ICON_SIZE_MENU
      self._start_menu = XdgMenu(self.get_xmenu(),
          icon_size = size, preload_icons = preload_icons.int_value,
          show_missing = show_missing.int_value,
          terminal = terminal.value,
          uri_handler = uri_handler.value,
          wrapper_dir = wrapper_dir.value,
          force_utf8_wrappers = force_utf8_wrappers.int_value,
          show_hidden = show_hidden.int_value)
      extra = []
      for xdg_menu_file in include_xdg_menus.value.split(','):
        xdg_menu_file_stripped = xdg_menu_file.strip()
        if not xdg_menu_file_stripped:
          continue
        xmenu = self.get_xmenu(xdg_menu_file_stripped)
        if xmenu is None:
          continue
        extra.append(XdgMenuItem(xmenu, xdg_menu_file_stripped,
          icon_size=size, preload_icons=preload_icons.int_value,
          show_missing=show_missing.int_value,
          terminal=terminal.value,
          uri_handler=uri_handler.value,
          wrapper_dir=wrapper_dir.value,
          force_utf8_wrappers=force_utf8_wrappers.int_value,
          show_hidden=show_hidden.int_value))
      if enable_rox_apps.int_value:
        extra.append(RoxMenuItem(self.get_rox_tree(),
          _("ROX Applications"),
          icon_size = size, preload_icons = preload_icons.int_value,
          show_missing = show_missing.int_value))
      if enable_logout.int_value:
        def find_rox_session(tree):
          try:
            rs = tree["ROX-Session"]
            if type(rs) != dict: return rs
          except KeyError: pass
          for n, p in tree.iteritems():
            if type(p) == dict:
              rs = find_rox_session(p)
              if not rs is None: return rs
          return None
        rox_session = find_rox_session(self.get_rox_tree())
        if rox_session:
          extra.append(RoxMenuItem(rox_session, _("Logout..."),
            icon_size = size, preload_icons = preload_icons.int_value,
            show_missing = show_missing.int_value))
        else:
          rox.alert(_("Unable to initialize the Logout menu item "
            "because ROX-Session is not found in the ROX Applications "
            "directories."))
      if len(extra) > 0:
        self._start_menu.append(g.SeparatorMenuItem())
        for i in extra:
          self._start_menu.append(i)
      self._start_menu.connect("deactivate", self._deactivate)
    return self._start_menu

  def refresh(self, *args):
    self._xmenu.clear()
    self._rox_tree = None
    self._start_menu = None
    xdg.Config.setIconTheme(
        g.settings_get_default().get_property("gtk-icon-theme-name"))
    self.get_start_menu()

  def popup_start_menu(self, *args, **kw):
    position_menu = kw.get('position_menu',
        self.position_menu)
    button = kw.get('button', 0)
    activate_time = kw.get('activate_time',
        g.get_current_event_time())
    self.set_state(g.STATE_SELECTED)
    m = self.get_start_menu()
    m.show_all()
    m.popup(None, None, position_menu,
        button, activate_time)
    return True

  def popup_start_menu_at_pointer(self, *args, **kw):
    def position_menu(menu):
      root = g.gdk.get_default_root_window()
      x, y, mods = root.get_pointer()
      return (x, y, True)
    kw['position_menu'] = position_menu
    return self.popup_start_menu(**kw)

  def _deactivate(self, *args):
    self.set_state(g.STATE_NORMAL)

  def _button_press(self, w, e):
    if e.type == g.gdk.BUTTON_PRESS and e.button == 1:
      return self.popup_start_menu(button = e.button)
    return False

  def is_vertical(self):
    try:
      pos = self.get_parent().socket.property_get(
          '_ROX_PANEL_MENU_POS', 'STRING', False)
      if pos:
        pos = pos[2]
        if pos: return pos[0] in ('L', 'R')
    except AttributeError: pass
    return False

  def position_menu(self, menu):
    try:
      x, y, z = self.get_parent().position_menu(menu)
    except AttributeError:
      try:
        x, y = self.get_parent().get_position()
      except AttributeError:
        x, y = (0, 0)
    sx, sy = self.window.get_origin()
    if self.flags() & g.NO_WINDOW:
      sx += self.allocation.x
      sy += self.allocation.y
    if self.is_vertical():
      y = sy
    else:
      x = sx
    return (x, y, True)

  def _options_changed(self):
    if (icon_size.has_changed or
        terminal.has_changed or
        uri_handler.has_changed or
        preload_icons.has_changed or
        show_hidden.has_changed or
        show_missing.has_changed or
        wrapper_dir.has_changed or
        force_utf8_wrappers.has_changed or
        app_dir_defs.has_changed or
        enable_rox_apps.has_changed or
        enable_logout.has_changed or
        include_xdg_menus.has_changed):
      self._start_menu = None
    if applet_image.has_changed:
      self._setup_image()
    if app_dir_defs.has_changed:
      self._rox_tree = None

  def quit(self, *args):
    try: self.get_parent().destroy()
    except AttributeError: self.destroy()

