import os
import sys
import shlex
import rox
import rox.fileutils
import rox.processes
from rox import g
import xdg.Menu
import xdg.DesktopEntry
import xdg.IconTheme
from i18n import _
from imagemenu import ImageMenu, ImageMenuItem
from dndsource import DndSourceWidget

def shlex_split(s):
  return map(lambda x: unicode(x, "UTF-8"),
      shlex.split(s.encode("UTF-8")))

class XdgMenu(ImageMenu):
  def __init__(self, xmenu, **kw):
    ImageMenu.__init__(self)
    dir = xmenu.getPath()
    for e in xmenu.getEntries(hidden = kw.get('show_hidden', False)):
      if isinstance(e, xdg.Menu.Menu):
        de = e
      elif isinstance(e, xdg.Menu.MenuEntry):
        de = e.DesktopEntry
      else: continue
      item = XdgMenuItem(de, dir, **kw)
      if isinstance(e, xdg.Menu.Menu) and \
          len(item.get_submenu().get_children()) == 0:
        item.destroy()
        continue
      item.set_tooltip_text(de.getComment())
      self.append(item)

class XdgMenuItem(ImageMenuItem, DndSourceWidget):
  def __init__(self, xentry, xdir, **kw):
    self.xentry = xentry
    self.xdir = xdir
    this_kw = kw
    if isinstance(xentry, xdg.Menu.Menu):
      this_kw = kw.copy()
      this_kw['missing_icon'] = g.STOCK_DIRECTORY
    ImageMenuItem.__init__(self, xentry.getName(), **this_kw)
    if isinstance(xentry, xdg.Menu.Menu):
      self.set_submenu(XdgMenu(xentry, **kw))
    else:
      self.terminal = kw['terminal']
      self.uri_handler = kw['uri_handler']
      self.connect("activate", self.run)
    self.wrapper_dir = kw['wrapper_dir']
    self.force_utf8_wrappers = kw['force_utf8_wrappers']
    if self.wrapper_dir:
      self.dnd_source_get_icon_path = self.get_icon_path
      self.dnd_source_enable()

  def get_icon_path(self):
    s, s = g.icon_size_lookup(self.icon_size)
    return xdg.IconTheme.getIconPath(
        self.xentry.getIcon(), s)

  def run(self, *args):
    path = None
    cmdv = None
    type = self.xentry.getType() 
    if type == "Application":
      e = self.get_exec_expanded()
      if e:
        cmdv = shlex_split(e)
        if self.xentry.getTerminal():
          cmdv[0:0] = shlex_split(self.terminal) + ["-e"]
        path = self.xentry.getPath()
    elif type == "Link":
      cmdv = shlex_split(
          self.uri_handler) + [self.xentry.getURL()]
    rox.processes.PipeThroughCommand(cmdv, None, None).start()

  def get_exec_shell(self):
    type = self.xentry.getType() 
    if type == "Application":
      cmd = self.get_exec_expanded()
      if cmd and self.xentry.getTerminal():
        cmd = self.terminal + " -e " + cmd
    elif type == "Link":
      cmd = self.uri_handler + " " + self.xentry.getURL()
    return cmd

  def get_exec_expanded(self):
    e = self.xentry.getExec()
    if e is None: return None
    icon = self.xentry.getIcon()
    cmd = ""
    i, n = 0, len(e)
    while i < n:
      c = e[i]; i += 1
      if c != "%":
        cmd += c
      else:
        c = e[i]; i = i+1
        if c == "%":
          cmd += c
        elif c == "i" and icon:
          cmd += "--icon \"%s\"" % icon
        elif c == "c":
          cmd += "\"%s\"" % self.xentry.getName()
    return cmd

  def get_wrapper_path(self):
    if self.xdir:
      p = ([self.wrapper_dir] + self.xdir.split("/") +
          [self.xentry.getName()])
    else: p = [self.wrapper_dir, self.xentry.getName()]
    path = apply(os.path.join, p)
    if self.force_utf8_wrappers:
      path = path.encode("UTF-8")
    return path

  def make_wrapper(self):
    path = self.get_wrapper_path()
    sig = "Created with XDG-Menu"
    if not isinstance(self.xentry, xdg.Menu.Menu):
      app_run = os.path.join(path, "AppRun")
      if os.access(app_run, os.X_OK):
        src = open(app_run)
        if src.read().find(sig) == -1: return path
        src.close()
    try: rox.fileutils.makedirs(path, 0755)
    except OSError: return path
    try:
      src = open(xdg.IconTheme.getIconPath(
        self.xentry.getIcon(), None))
    except: pass
    else:
      dest = open(os.path.join(path, ".DirIcon"), "w")
      while True:
        buf = src.read(2048)
        if not buf: break
        dest.write(buf)
      src.close()
      dest.close()
    if isinstance(self.xentry, xdg.Menu.Menu):
      for c in self.get_submenu().get_children():
        if isinstance(c, XdgMenuItem):
          c.make_wrapper()
    else:
      dest = open(app_run, "w")
      dest.write("#!/bin/sh\n# %s\n%s \"$@\"\n" %
          (sig, self.get_exec_shell()))
      dest.close()
      os.chmod(app_run, 0755)
    return path

  def dnd_source_get_uri(self):
    if (isinstance(self.xentry, xdg.Menu.Menu) or 
        self.xentry.getType() == "Application"):
      return "file://" + rox.our_host_name() + rox.escape(
          self.make_wrapper())
    elif self.xentry.getType() == "Link":
      return self.xentry.getURL()
    return ""
